/* ScummVM - Graphic Adventure Engine
 *
 * ScummVM is the legal property of its developers, whose names
 * are too numerous to list here. Please refer to the COPYRIGHT
 * file distributed with this source distribution.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef HODJNPODJ_POKER_RESOURCE_H
#define HODJNPODJ_POKER_RESOURCE_H

namespace Bagel {
namespace HodjNPodj {
namespace Poker {

#define IDI_POKER                       2
#define IDB_BITMAP1                     154
#define IDB_BITMAP2                     155
#define IDB_BITMAP3                     156
#define IDB_BITMAP4                     157
#define IDB_BITMAP5                     158
#define IDB_BITMAP6                     159
#define IDB_BITMAP7                     160
#define IDB_BITMAP8                     161
#define IDB_BITMAP9                     162
#define IDB_BITMAP10                    163
#define IDB_BITMAP11                    164
#define IDB_BITMAP12                    165
#define IDB_BITMAP13                    166
#define IDB_BITMAP14                    167
#define IDB_BITMAP15                    168
#define IDB_BITMAP16                    169
#define IDB_BITMAP17                    170
#define IDB_BITMAP18                    171
#define IDB_BITMAP19                    172
#define IDB_BITMAP20                    173
#define IDB_BITMAP21                    174
#define IDB_BITMAP22                    175
#define IDB_BITMAP23                    176
#define IDB_BITMAP24                    177
#define IDB_BITMAP25                    178
#define IDB_BITMAP26                    179
#define IDB_BITMAP27                    180
#define IDB_BITMAP28                    181
#define IDB_BITMAP29                    182
#define IDB_BITMAP30                    183
#define IDB_BITMAP31                    184
#define IDB_BITMAP32                    185
#define IDB_BITMAP33                    186
#define IDB_BITMAP34                    187
#define IDB_BITMAP35                    188
#define IDB_BITMAP36                    189
#define IDB_BITMAP37                    190
#define IDB_BITMAP38                    191
#define IDB_BITMAP39                    192
#define IDB_BITMAP40                    193
#define IDB_BITMAP41                    194
#define IDB_BITMAP42                    195
#define IDB_BITMAP43                    196
#define IDB_BITMAP44                    197
#define IDB_BITMAP45                    198
#define IDB_BITMAP46                    199
#define IDB_BITMAP47                    200
#define IDB_BITMAP48                    201
#define IDB_BITMAP49                    202
#define IDB_BITMAP50                    203
#define IDB_BITMAP51                    204
#define IDB_BITMAP52                    205
#define IDD_SETAMOUNT                   235
#define IDD_DIALOG1                     236
#define IDD_WINRATIO                    237
#define IDD_DIALOG2                     238
#define IDD_OPTIONS                     239
#define IDB_GRABAGE                     242
#define IDB_BET1000UP                   248
#define IDB_BET100UP                    249
#define IDB_BET10UP                     250
#define IDB_BET1DISABLED                251
#define IDB_BET1DOWN                    252
#define IDB_BET1UP                      253
#define IDB_BET25UP                     254
#define IDB_BET5UP                      255
#define IDB_BETALLUP                    256
#define IDB_CLEARUP                     257
#define IDB_DEALDISABLED                258
#define IDB_DEALDOWN                    259
#define IDB_DEALUP                      260
#define IDB_DRAWDIABLED                 261
#define IDB_DRAWDOWN                    262
#define IDB_DRAWUP                      263
#define IDB_HOLDLIT                     264
#define IDB_HOLDDARK                    265
#define IDB_HOLDDISABLED                266
#define IDB_HOLDDOWN                    267
#define IDB_LSCROLLB                    268
#define IDB_LSCROLLM                    269
#define IDB_LSCROLLT                    270
#define IDB_POKER                       271
#define IDB_SSCROLL                     272
#define IDD_USERWON                     273
#define IDB_BITMAP53                    274
#define IDB_JOKER                       274
#define IDB_CARDBACK                    275
#define IDB_HOLD1DISABLED               276
#define IDB_HOLD1DOWN                   277
#define IDB_HOLD1HALF                   278
#define IDB_HOLD1LIT                    279
#define IDB_HOLD2DISABLED               280
#define IDB_HOLD2DOWN                   281
#define IDB_HOLD2HALF                   282
#define IDB_HOLD2LIT                    283
#define IDB_HOLD3DISABLED               284
#define IDB_HOLD3DOWN                   285
#define IDB_HOLD3HALF                   286
#define IDB_HOLD3LIT                    287
#define IDB_HOLD4DISABLED               288
#define IDB_HOLD4DOWN                   289
#define IDB_HOLD4HALF                   290
#define IDB_HOLD4LIT                    291
#define IDB_HOLD5DISABLED               292
#define IDB_HOLD5DOWN                   293
#define IDB_HOLD5HALF                   294
#define IDB_HOLD5LIT                    295
#define IDB_OPTIONSCROLL                296
#define IDB_DEALGRAY                    298
#define IDD_USERAMOUNT                  301
#define IDB_DRAWGRAY                    301
#define IDC_USERAMOUNTDISPLAY           302
#define IDC_SETAMT_BAR                  303
#define IDB_BET100DOWN                  303
#define IDC_POPO                        304
#define ID_POPO                         304
#define IDB_BET100GRAY                  304
#define IDC_KLINGON                     305
#define ID_KLINGON                      305
#define IDC_VEGAS                       306
#define ID_VEGAS                        306
#define IDB_BET10DOWN                   306
#define IDC_MARTIAN                     307
#define ID_MARTIAN                      307
#define IDB_BET10GRAY                   307
#define IDC_KUWAITI                     308
#define ID_KUWAITI                      308
#define IDC_PAIR                        309
#define IDC_PAIR2                       310
#define IDB_BET1GRAY                    310
#define ID_SETUSERAMT                   311
#define IDB_BET1000DOWN                 311
#define IDC_3OFAKIND                    312
#define IDB_BET1000GRAY                 312
#define IDC_RULES                       313
#define ID_SETPAYOFFS                   314
#define IDC_TWOPAIR                     315
#define IDB_BET25DOWN                   315
#define IDC_SETSTARTAMOUNT              316
#define IDB_BET25GRAY                   316
#define ID_TOGGLESOUND                  317
#define IDC_STRAIGHT                    318
#define IDB_BET5DOWN                    318
#define IDC_FLUSH                       319
#define IDB_BET5GRAY                    319
#define IDC_QUIT                        320
#define IDC_FULLHOUSE                   321
#define IDB_BETALLDOWN                  321
#define IDC_4OFAKIND                    322
#define IDB_BETALLGRAY                  322
#define IDC_STRAIGHTFLUSH               323
#define IDC_ROYALFLUSH                  324
#define IDC_SETPAYOFFS                  325
#define IDB_CLEARBETDOWN                327
#define IDB_CLEARBETGRAY                328
#define IDB_CLEARBETUP                  329
#define IDD_RULES_DIALOG                900
#define IDC_RULES_OKAY                  900
#define IDC_CANCEL                      1019
#define IDC_OPTIONOK                    1020
#define ID_CANCEL                       1021
#define ID_CANCEL2                      1021
#define ID_CANCEL_VIEW                  1027

} // namespace Poker
} // namespace HodjNPodj
} // namespace Bagel

#endif
