/* ScummVM - Graphic Adventure Engine
 *
 * ScummVM is the legal property of its developers, whose names
 * are too numerous to list here. Please refer to the COPYRIGHT
 * file distributed with this source distribution.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef BAGEL_METAGAME_ZOOM_RESOURCE_H
#define BAGEL_METAGAME_ZOOM_RESOURCE_H

namespace Bagel {
namespace HodjNPodj {
namespace Metagame {
namespace Zoom {

#define IDI_WORDSRCH                    2
#define IDC_OPTIONS                     202
#define IDB_LSCROLLB                    210
#define IDB_LSCROLLM                    211
#define IDB_LSCROLLT                    212
#define IDB_SSCROLL                     213
#define IDB_OPTIONSCROLL                220
#define IDC_RULES                       230
#define IDC_CANCEL                      250
#define IDC_QUIT                        251
#define IDB_PODJRIGHT                   330
#define IDD_MESSAGEBOX                  333
#define IDB_OPTIONG                     359
#define IDB_OPTIONU                     360
#define IDB_LETTERA                     361
#define IDB_LETTERB                     362
#define IDB_LETTERC                     363
#define IDB_LETTERD                     364
#define IDB_LETTERE                     365
#define IDB_LETTERF                     366
#define IDB_LETTERG                     367
#define IDB_LETTERH                     368
#define IDB_LETTERI                     369
#define IDB_LETTERJ                     370
#define IDB_LETTERK                     371
#define IDB_LETTERL                     372
#define IDB_LETTERM                     373
#define IDB_LETTERN                     374
#define IDB_LETTERO                     375
#define IDB_LETTERP                     376
#define IDB_LETTERQ                     377
#define IDB_LETTERR                     378
#define IDB_LETTERS                     379
#define IDB_LETTERT                     380
#define IDB_LETTERU                     381
#define IDB_LETTERV                     382
#define IDB_LETTERW                     383
#define IDB_LETTERX                     384
#define IDB_LETTERY                     385
#define IDB_LETTERZ                     386
#define IDB_LETTERA_D                   387
#define IDB_LETTERB_D                   388
#define IDB_LETTERC_D                   389
#define IDB_LETTERD_D                   390
#define IDB_LETTERE_D                   391
#define IDB_LETTERF_D                   392
#define IDB_LETTERG_D                   393
#define IDB_LETTERH_D                   394
#define IDB_LETTERI_D                   395
#define IDB_LETTERJ_D                   396
#define IDB_LETTERK_D                   397
#define IDB_LETTERL_D                   398
#define IDB_LETTERM_D                   399
#define IDB_LETTERN_D                   400
#define IDB_LETTERO_D                   401
#define IDB_LETTERP_D                   402
#define IDB_LETTERQ_D                   403
#define IDB_LETTERR_D                   404
#define IDB_LETTERS_D                   405
#define IDB_LETTERT_D                   406
#define IDB_LETTERU_D                   407
#define IDB_LETTERV_D                   408
#define IDB_LETTERW_D                   409
#define IDB_LETTERX_D                   410
#define IDB_LETTERY_D                   411
#define IDB_LETTERZ_D                   412
#define IDC_CURSOR1                     415
#define IDB_BEAVER1                     419
#define IDB_BEAVER2                     420
#define IDB_BEAVER3                     421
#define IDB_BEAVER4                     422
#define IDB_BEAVER5                     423
#define IDB_BEAVER6                     424
#define IDB_BEAVER7                     425
#define IDD_MINIOPTIONS_DIALOG          500
#define IDC_MINI_CANCEL                 501
#define IDC_MINI_OK                     502
#define IDC_GAMELEVEL                   503
#define IDC_MINI_WLTOGGLE               503
#define IDC_NUMBEROFLIVES               504
#define IDC_GAMETIME                    504
#define IDC_BEAVERDUR                   505

} // namespace Zoom
} // namespace Metagame
} // namespace HodjNPodj
} // namespace Bagel

#endif
