/* ScummVM - Graphic Adventure Engine
 *
 * ScummVM is the legal property of its developers, whose names
 * are too numerous to list here. Please refer to the COPYRIGHT
 * file distributed with this source distribution.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef BAGEL_MFC_GLOBAL_MENURES_H
#define BAGEL_MFC_GLOBAL_MENURES_H

#define IDC_OPTIONS_ARROWUP 503
#define IDC_OPTIONS_ARROWDN 504

#define ID_OPTIONS_CODES    450

// obsolete
//#define IDR_BITMAP_SCROLL   451

#define IDR_OPTION_SCROLL   452
#define IDD_OPTIONS_DIALOG  453

#define IDC_OPTIONS_RETURN  454
#define IDC_OPTIONS_QUIT    455
#define IDC_OPTIONS_RULES   456
#define IDC_OPTIONS_NEWGAME 457
#define IDC_OPTIONS_OPTIONS 458
#define IDC_OPTIONS_AUDIO   459
#define IDC_OPTIONS_HYPE    999

#define IDB_SCROLBTN        460
#define SCROLLUP            461
#define SCROLLDOWN          462

#define IDD_AUDIOCFG        463

#endif
