begin_section("SCUMM");
	add_person("Torbj&ouml;rn Andersson", "eriktorbjorn", "");
	add_person("Andrea Boscarino", "AndywinXp", "Digital iMUSE, SMUSH audio, GUI");
	add_person("James Brown", "ender", "(retired)");
	add_person("Jonathan Gray", "khalek", "(retired)");
	add_person("Vincent Hamm", "yaz0r", "(retired)");
	add_person("Max Horn", "Fingolfin", "(retired)");
	add_person("Travis Howell", "Kirben", "");
	add_person("Pawe&#322; Ko&#322;odziejski", "aquadran", "Codecs, iMUSE, Smush, etc.");
	add_person("Gregory Montoir", "cyx", "(retired)");
	add_person("Eugene Sandulenko", "sev", "FT INSANE, MM NES, MM C64, game detection, Herc/CGA");
	add_person("Ludvig Strigeus", "ludde", "(retired)");
end_section();

begin_section("SCUMM HE");
	add_person("Andrea Boscarino", "AndywinXp", "Digital audio, WIZ graphics system");
	add_person("Jonathan Gray", "khalek", "(retired)");
	add_person("Travis Howell", "Kirben", "");
	add_person("Gregory Montoir", "cyx", "(retired)");
	add_person("Eugene Sandulenko", "sev", "");
end_section();
