begin_section("MM (Xeen)");
	add_person("Paul Gilbert", "dreammaster", "");
	add_person("Benoit Pierre", "benoit-pierre", "");
	add_person("", "TheDrakeRaider", "(M&M1 Gfx Mod)");
	add_person("David Goldsmith", "WizardStan", "(Xeen analysis)");
	add_person("Matt Taylor", "", "(Xeen analysis)");
end_section();
