begin_section("Trecision");
	add_person("Daniel Albano", "SupSuper", "");
	add_person("Arnaud Boutonn&eacute;", "Strangerke", "");
	add_person("Thomas Fach-Pedersen", "madmoose", "Smacker video support");
	add_person("Filippos Karapetis", "bluegr", "");
end_section();
