/* ScummVM - Graphic Adventure Engine
 *
 * ScummVM is the legal property of its developers, whose names
 * are too numerous to list here. Please refer to the COPYRIGHT
 * file distributed with this source distribution.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef ZVISION_REGION_NODE_H
#define ZVISION_REGION_NODE_H

#include "graphics/surface.h"
#include "zvision/graphics/graphics_effect.h"
#include "zvision/scripting/scripting_effect.h"

namespace ZVision {

class ZVision;

class RegionNode : public ScriptingEffect {
public:
	RegionNode(ZVision *engine, uint32 key, GraphicsEffect *effect, uint32 delay);
	~RegionNode() override;

	/**
	 * Decrement the timer by the delta time. If the timer is finished, set the status
	 * in _globalState and let this node be deleted
	 *
	 * @param deltaTimeInMillis    The number of milliseconds that have passed since last frame
	 * @return                     If true, the node can be deleted after process() finishes
	 */
	bool process(uint32 deltaTimeInMillis) override;

private:
	int32 _timeLeft;
	uint32 _delay;
	GraphicsEffect *_effect;
};

} // End of namespace ZVision

#endif
