/* ScummVM - Graphic Adventure Engine
 *
 * ScummVM is the legal property of its developers, whose names
 * are too numerous to list here. Please refer to the COPYRIGHT
 * file distributed with this source distribution.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef TETRAEDGE_TE_TE_FRAME_ANIM_H
#define TETRAEDGE_TE_TE_FRAME_ANIM_H

#include "tetraedge/te/te_signal.h"
#include "tetraedge/te/te_animation.h"

namespace Tetraedge {

class TeFrameAnim : public TeAnimation {
public:
	TeFrameAnim();
	virtual ~TeFrameAnim() {}

	void update(double millis) override;

	TeSignal0Param &frameChangedSignal() { return _frameChangedSignal; };

	void setFrameRate(float rate) { _frameRate = rate; }
	void setNbFrames(int frames) { _nbFrames = frames; }
	void setLoopCount(int count) { _loopCount = count; }
	void setReversed(bool reverse) { _reversed = reverse; }

	void setStartTime(double start) { _startTime = start; }
	void setEndTime(double end) { _endTime = end; }

	int lastFrameShown() const { return _lastFrameShown; }

private:
	float _frameRate;
	// TODO: Isn't this the same as TeAnimation::_repeatCount??
	int _loopCount;
	int _nbFrames;
	int _numFramesToShow;

	bool _reversed;

	int _lastFrameShown;
	int _minFrame;

	double _startTime;
	double _endTime;

	TeSignal0Param _frameChangedSignal;
};

} // end namespace Tetraedge

#endif // TETRAEDGE_TE_TE_FRAME_ANIM_H
