// This file was generated by credits.pl. Do not edit by hand!
static const char *const credits[] = {
"C1""ScummVM Team",
"",
"C1""Project Leaders",
"C0""Pawe\305\202 Ko\305\202odziejski",
"C0""Eugene Sandulenko",
"C0""Einar Johan T. S\303\270m\303\245en",
"C0""Lothar Serra Mari",
"",
"C1""PR Office",
"C0""Arnaud Boutonn\303\251",
"C2""Public Relations Officer, Project Administrator",
"C0""Eugene Sandulenko",
"C2""Project Leader",
"",
"C1""Retired Project Leaders",
"C0""James Brown",
"C0""Vincent Hamm",
"C2""ScummVM co-founder, Original Cruise/CinE author",
"C0""Max Horn",
"C0""Ludvig Strigeus",
"C2""Original ScummVM and SimonVM author",
"",
"C1""Engine Teams",
"C1""Access",
"C0""Arnaud Boutonn\303\251",
"C0""Paul Gilbert",
"C0""Matthew Duggan",
"",
"C1""ADL",
"C0""Walter van Niftrik",
"",
"C1""AGI",
"C0""Stuart George",
"C0""Matthew Hoops",
"C2""(retired)",
"C0""Filippos Karapetis",
"C0""Martin Kiewitz",
"C0""Pawe\305\202 Ko\305\202odziejski",
"C0""Walter van Niftrik",
"C0""Kari Salminen",
"C0""Eugene Sandulenko",
"C0""David Symonds",
"C2""(retired)",
"",
"C1""AGOS",
"C0""Torbj\303\266rn Andersson",
"C0""Paul Gilbert",
"C0""Travis Howell",
"C0""Oliver Kiehl",
"C2""(retired)",
"C0""Ludvig Strigeus",
"C2""(retired)",
"",
"C1""AGS",
"C0""Paul Gilbert",
"C0""Thierry Crozat",
"C0""Walter Agazzi",
"C2""Maintenance, Game archival",
"C0""Chris Jones",
"C2""Creator",
"C0""Alan Van Drake",
"C2""AGS",
"C0""Benjamin Penney",
"C2""AGS",
"C0""Benoit Pierre",
"C2""AGS",
"C0""Bernhard Rosenkraenzer",
"C2""AGS",
"C0""Cristian Morales Vega",
"C2""AGS",
"C0""Edward Rudd",
"C2""AGS",
"C0""Erico Vieira Porto",
"C2""AGS",
"C0""Ferdinand Thiessen",
"C2""AGS",
"C0""Francesco Ariis",
"C2""AGS",
"C0""Gilad Shaham",
"C2""AGS",
"C0""Ivan Mogilko",
"C2""AGS",
"C0""Janet Gilbert",
"C2""AGS",
"C0""Jochen Schleu",
"C2""AGS",
"C0""Joe Lee",
"C2""AGS",
"C0""John Steele Scott",
"C2""AGS",
"C0""Martin Sedlak",
"C2""AGS",
"C0""Matthew Gambrell",
"C2""AGS",
"C0""Michael Rittenhouse",
"C2""AGS",
"C0""Morgan Willcock",
"C2""AGS",
"C0""Nick Sonneveld",
"C2""AGS",
"C0""Ori Avtalion",
"C2""AGS",
"C0""Paul Wilkinson",
"C2""AGS",
"C0""Per Olav Flaten",
"C2""AGS",
"C0""Piotr Wieczorek",
"C2""AGS",
"C0""Ryan O'Connor",
"C2""AGS",
"C0""Scott Baker",
"C2""AGS",
"C0""Shane Stevens",
"C2""AGS",
"C0""Shawn R. Walker",
"C2""AGS",
"C0""Stefano Collavini",
"C2""AGS",
"C0""Steve McCrea",
"C2""AGS",
"C0""Steven Poulton",
"C2""AGS",
"C0""Sunit Das",
"C2""AGS",
"C0""Tobias Hansen",
"C2""AGS",
"C0""Tom Vandepoele",
"C2""AGS",
"C0""Tzach Shabtay",
"C2""AGS",
"C0""rofl0r",
"C2""AGS",
"C0""Berian Williams",
"C2""AgsCreditz",
"",
"C1""Alcachofa",
"C0""Hermann Noll",
"",
"C1""Asylum",
"C0""Alex Bevilacqua",
"C0""Alex Fontoura",
"C0""Alexander Panov",
"C0""Benjamin Haisch",
"C0""Filippos Karapetis",
"C0""Joseph Davies",
"C0""Julien Templier",
"",
"C1""Avalanche",
"C0""Peter Bozs\303\263",
"C0""Arnaud Boutonn\303\251",
"",
"C1""awe",
"C0""Gregory Montoir",
"C0""Paul Gilbert",
"",
"C1""Bagel",
"C0""Arnaud Boutonn\303\251",
"C0""Daniel Albano",
"C0""Eugene Sandulenko",
"C0""Paul Gilbert",
"",
"C1""BBVS",
"C0""Benjamin Haisch",
"",
"C1""Blade Runner",
"C0""Thanasis Antoniou",
"C0""Thomas Fach-Pedersen",
"C0""Peter Kohaut",
"C0""Eugene Sandulenko",
"",
"C1""Buried",
"C0""Matthew Hoops",
"",
"C1""CGE",
"C0""Arnaud Boutonn\303\251",
"C0""Paul Gilbert",
"",
"C1""CGE2",
"C0""Peter Bozs\303\263",
"C0""Arnaud Boutonn\303\251",
"C0""Paul Gilbert",
"",
"C1""Chamber",
"C0""Retro-Junk;",
"C0""Eugene Sandulenko",
"",
"C1""Chewy",
"C0""Paul Gilbert",
"C0""Eugene Sandulenko",
"C0""Arnaud Boutonn\303\251",
"C0""Filippos Karapetis",
"",
"C1""Cine",
"C0""Vincent Hamm",
"C2""(retired)",
"C0""Pawe\305\202 Ko\305\202odziejski",
"C0""Gregory Montoir",
"C2""(retired)",
"C0""Kari Salminen",
"C0""Eugene Sandulenko",
"",
"C1""Composer",
"C0""Alyssa Milburn",
"",
"C1""CRAB",
"C0""Kartik Agarwala",
"",
"C1""CruisE",
"C0""Paul Gilbert",
"C0""Vincent Hamm",
"C2""(retired)",
"",
"C1""Cryo",
"C0""Arnaud Boutonn\303\251",
"C0""Filippos Karapetis",
"C0""Retro-Junk;",
"C0""Eugene Sandulenko",
"",
"C1""Cryomni3D",
"C0""Philippe Valembois",
"",
"C1""Darkseed",
"C0""Eric Fry",
"",
"C1""DGDS",
"C0""Matthew Duggan",
"C0""Filippos Karapetis",
"C0""Vasco Alexandre da Silva Costa",
"",
"C1""Director",
"C0""Eugene Sandulenko",
"C0""Crane Yang",
"C2""GSoC Student",
"C0""Dmitry Iskrich",
"C0""Deborah Servilla",
"C2""GSoC Student",
"C0""Nathanael Gentry",
"C2""GSoC Student",
"C0""Roland van Laar",
"C0""Scott Percival",
"C0""Steven Hoefel",
"C0""Tobia Tesan",
"C0""Misty De M\303\251o",
"",
"C1""DM",
"C0""Arnaud Boutonn\303\251",
"C0""Bendeg\303\272z Nagy",
"",
"C1""Draci",
"C0""Denis Kasak",
"C0""Robert \305\240palek",
"",
"C1""Dragons",
"C0""Eric Fry",
"C0""Benjamin Haisch",
"C2""Actor pathfinding",
"C0""\303\201ngel Eduardo Garc\303\255a Hern\303\241ndez",
"C2""Help with reverse engineering",
"",
"C1""Drascula",
"C0""Filippos Karapetis",
"C0""Pawe\305\202 Ko\305\202odziejski",
"C0""Thierry Crozat",
"",
"C1""DreamWeb",
"C0""Torbj\303\266rn Andersson",
"C0""Bertrand Augereau",
"C0""Filippos Karapetis",
"C0""Vladimir Menshakov",
"C0""Willem Jan Palenstijn",
"",
"C1""Efh",
"C0""Arnaud Boutonn\303\251",
"",
"C1""Freescape",
"C0""Chris Allen",
"C2""Sound engine programming",
"C0""Gustavo Grieco",
"",
"C1""Glk",
"C0""Paul Gilbert",
"C0""Tor Andersson",
"C2""GarGlk library",
"C0""Stefan Jokisch",
"C2""Frotz interpreter",
"C0""Andrew Plotkin",
"C2""Glulxe interpreter",
"C0""Alan Cox",
"C2""ScottFree interpreter",
"C0""Michael J. Roberts",
"C2""TADS interpreter",
"C0""Avijeet Maurya",
"C2""Scott Sub-engine",
"",
"C1""Gnap",
"C0""Arnaud Boutonn\303\251",
"C0""Benjamin Haisch",
"",
"C1""Gob",
"C0""Torbj\303\266rn Andersson",
"C0""Arnaud Boutonn\303\251",
"C0""Simon Delamarre",
"C0""Sven Hesse",
"C0""Eugene Sandulenko",
"",
"C1""Got",
"C0""Paul Gilbert",
"C0""Arnaud Boutonn\303\251",
"C0""Coen Rampen",
"",
"C1""Griffon",
"C0""Eugene Sandulenko",
"",
"C1""Grim",
"C0""James Brown",
"C2""Grim (retired)",
"C0""Giulio Camuffo",
"C2""Grim (retired)",
"C0""Daniel Schepler",
"C2""Initial engine contributor",
"C0""Dries Harnie",
"C2""EMI",
"C0""Pawe\305\202 Ko\305\202odziejski",
"C2""Grim",
"C0""Christian Krause",
"C2""EMI (retired)",
"C0""Einar Johan T. S\303\270m\303\245en",
"C2""Grim, EMI",
"C0""Joel Teichroeb",
"C2""EMI",
"C0""Joni V\303\244h\303\244m\303\244ki",
"C2""EMI (retired)",
"",
"C1""Groovie",
"C0""Henry Bush",
"C0""Ray Carro",
"C0""Scott Thomas",
"C0""Jordi Vilalta Prat",
"",
"C1""Hades Challenge",
"C0""Vladimir Serbinenko/Google",
"",
"C1""HDB",
"C0""Eugene Sandulenko",
"C0""Nipun Garg",
"C2""GSoC student",
"",
"C1""Hopkins",
"C0""Arnaud Boutonn\303\251",
"C0""Paul Gilbert",
"",
"C1""Hpl1",
"C0""Emanuele Grisenti",
"",
"C1""Hugo",
"C0""Arnaud Boutonn\303\251",
"C0""Oystein Eftevaag",
"C0""Eugene Sandulenko",
"",
"C1""Hypno",
"C0""Gustavo Grieco",
"",
"C1""ICB",
"C0""Pawe\305\202 Ko\305\202odziejski",
"C0""Joost Peters",
"C0""Einar Johan T. S\303\270m\303\245en",
"",
"C1""Illusions",
"C0""Benjamin Haisch",
"C0""Eric Fry",
"",
"C1""Immortal",
"C0""Michael Hayman",
"",
"C1""Kingdom",
"C0""Arnaud Boutonn\303\251",
"C0""Thomas Fach-Pedersen",
"C0""Hein-Pieter van Braam-Stewart",
"",
"C1""Kyra",
"C0""Torbj\303\266rn Andersson",
"C2""VQA Player",
"C0""Oystein Eftevaag",
"C0""Florian Kagerer",
"C0""Gregory Montoir",
"C2""(retired)",
"C0""Johannes Schickel",
"C2""(retired)",
"",
"C1""Lab",
"C0""Arnaud Boutonn\303\251",
"C0""Filippos Karapetis",
"C0""Willem Jan Palenstijn",
"C0""Eugene Sandulenko",
"",
"C1""Lastexpress",
"C0""Andrea Boscarino",
"C2""Engine rewrite",
"C0""Matthew Hoops",
"C2""(retired)",
"C0""Jordi Vilalta Prat",
"C0""Julien Templier",
"",
"C1""Lilliput",
"C0""Arnaud Boutonn\303\251",
"",
"C1""Lure",
"C0""Paul Gilbert",
"",
"C1""M4",
"C0""Paul Gilbert",
"C0""Arnaud Boutonn\303\251",
"",
"C1""MacVenture",
"C0""Borja Lorente",
"C2""GSoC student",
"",
"C1""MADE",
"C0""Benjamin Haisch",
"C0""Filippos Karapetis",
"",
"C1""MADS",
"C0""Arnaud Boutonn\303\251",
"C0""Paul Gilbert",
"C0""Filippos Karapetis",
"",
"C1""Media Station",
"C0""Nathanael Gentry",
"",
"C1""MM (Xeen)",
"C0""Paul Gilbert",
"C0""Benoit Pierre",
"C0""TheDrakeRaider",
"C2""(M&M1 Gfx Mod)",
"C0""David Goldsmith",
"C2""(Xeen analysis)",
"C0""Matt Taylor",
"C2""(Xeen analysis)",
"",
"C1""Mohawk",
"C0""Bastien Bouclet",
"C0""Matthew Hoops",
"C2""(retired)",
"C0""Filippos Karapetis",
"C0""Alyssa Milburn",
"C0""Eugene Sandulenko",
"C0""David Turner",
"C0""David Fioramonti",
"",
"C1""Mortevielle",
"C0""Arnaud Boutonn\303\251",
"C0""Paul Gilbert",
"",
"C1""mTropolis",
"C0""Eric Lasota",
"",
"C1""MutationOfJB",
"C0""\304\275ubom\303\255r Rem\303\241k",
"C0""Miroslav Rem\303\241k",
"",
"C1""Myst 3",
"C0""Bastien Bouclet",
"",
"C1""Nancy",
"C0""Kaloyan Chehlarski",
"C0""Walter van Niftrik",
"",
"C1""Neverhood",
"C0""Benjamin Haisch",
"C0""Filippos Karapetis",
"",
"C1""NGI",
"C0""Eugene Sandulenko",
"",
"C1""Parallaction",
"C0""peres",
"",
"C1""Pegasus",
"C0""Matthew Hoops",
"C2""(retired)",
"",
"C1""Petka",
"C0""Andrei Prykhodko",
"C0""Eugene Sandulenko",
"",
"C1""Pink",
"C0""Andrei Prykhodko",
"C0""Eugene Sandulenko",
"",
"C1""Plumbers",
"C0""Retro-Junk;",
"",
"C1""Prince",
"C0""Eugene Sandulenko",
"C0""\305\201ukasz W\304\205tka",
"C0""Kamil Zbr\303\263g",
"",
"C1""Private",
"C0""Gustavo Grieco",
"",
"C1""QDEngine",
"C0""Eugene Sandulenko",
"C0""Kunal Tiwari",
"",
"C1""Queen",
"C0""David Eriksson",
"C2""(retired)",
"C0""Gregory Montoir",
"C2""(retired)",
"C0""Joost Peters",
"",
"C1""SAGA",
"C0""Torbj\303\266rn Andersson",
"C0""Daniel Balsom",
"C2""Original engine reimplementation author (retired)",
"C0""Filippos Karapetis",
"C0""Andrew Kurushin",
"C0""Eugene Sandulenko",
"",
"C1""SAGA2",
"C0""Yuri Guimaraes",
"C2""GSoC Student",
"C0""Eugene Sandulenko",
"",
"C1""SCI",
"C0""Chris Benshoof",
"C0""Greg Frieger",
"C0""Paul Gilbert",
"C0""Max Horn",
"C2""(retired)",
"C0""Filippos Karapetis",
"C0""Martin Kiewitz",
"C0""Walter van Niftrik",
"C0""Willem Jan Palenstijn",
"C0""Jordi Vilalta Prat",
"C0""Lars Skovlund",
"C0""Colin Snover",
"",
"C1""SCUMM",
"C0""Torbj\303\266rn Andersson",
"C0""Andrea Boscarino",
"C2""Digital iMUSE, SMUSH audio, GUI",
"C0""James Brown",
"C2""(retired)",
"C0""Jonathan Gray",
"C2""(retired)",
"C0""Vincent Hamm",
"C2""(retired)",
"C0""Max Horn",
"C2""(retired)",
"C0""Travis Howell",
"C0""Pawe\305\202 Ko\305\202odziejski",
"C2""Codecs, iMUSE, Smush, etc.",
"C0""Gregory Montoir",
"C2""(retired)",
"C0""Eugene Sandulenko",
"C2""FT INSANE, MM NES, MM C64, game detection, Herc/CGA",
"C0""Ludvig Strigeus",
"C2""(retired)",
"",
"C1""SCUMM HE",
"C0""Andrea Boscarino",
"C2""Digital audio, WIZ graphics system",
"C0""Jonathan Gray",
"C2""(retired)",
"C0""Travis Howell",
"C0""Gregory Montoir",
"C2""(retired)",
"C0""Eugene Sandulenko",
"",
"C1""Sherlock",
"C0""Paul Gilbert",
"C0""Martin Kiewitz",
"",
"C1""Sky",
"C0""Robert G\303\266ffringmann",
"C2""(retired)",
"C0""Oliver Kiehl",
"C2""(retired)",
"C0""Joost Peters",
"",
"C1""SLUDGE",
"C0""Eugene Sandulenko",
"C0""Simei Yin",
"C2""GSoC Student",
"",
"C1""Stark",
"C0""Bastien Bouclet",
"C0""Einar Johan T. S\303\270m\303\245en",
"C0""Liu Zhaosong",
"",
"C1""Star Trek",
"C0""Matthew Hoops",
"C2""(retired)",
"C0""Filippos Karapetis",
"C0""Matthew Stewart",
"C2""GSoC Student",
"",
"C1""Supernova",
"C0""Joseph-Eugene Winzer",
"C0""Jarom\303\255r Wysoglad",
"C0""Thierry Crozat",
"",
"C1""Sword1",
"C0""Fabio Battaglia",
"C2""PSX version support",
"C0""Andrea Boscarino",
"C2""Palette fading, menus, audio",
"C0""Thierry Crozat",
"C2""Mac version support",
"C0""Robert G\303\266ffringmann",
"C2""(retired)",
"",
"C1""Sword2",
"C0""Torbj\303\266rn Andersson",
"C0""Fabio Battaglia",
"C2""PSX version support",
"C0""Jonathan Gray",
"C2""(retired)",
"",
"C1""Sword2.5",
"C0""Torbj\303\266rn Andersson",
"C0""Paul Gilbert",
"C0""Max Horn",
"C2""(retired)",
"C0""Filippos Karapetis",
"C0""Eugene Sandulenko",
"",
"C1""TeenAgent",
"C0""Robert Megone",
"C2""Help with callback rewriting",
"C0""Vladimir Menshakov",
"",
"C1""Tetraedge",
"C0""Matthew Duggan",
"",
"C1""Tinsel",
"C0""Torbj\303\266rn Andersson",
"C0""Fabio Battaglia",
"C2""PSX version support",
"C0""Paul Gilbert",
"C0""Sven Hesse",
"C0""Max Horn",
"C2""(retired)",
"C0""Filippos Karapetis",
"C0""Joost Peters",
"",
"C1""Titanic",
"C0""David Fioramonti",
"C0""Paul Gilbert",
"C0""Colin Snover",
"",
"C1""Toltecs",
"C0""Benjamin Haisch",
"C0""Filippos Karapetis",
"",
"C1""Tony",
"C0""Arnaud Boutonn\303\251",
"C0""Paul Gilbert",
"C0""Alyssa Milburn",
"",
"C1""Toon",
"C0""Sylvain Dupont",
"",
"C1""Tot",
"C0""Gabriel Sanmartín",
"",
"C1""Touch\303\251",
"C0""Gregory Montoir",
"C2""(retired)",
"",
"C1""Trecision",
"C0""Daniel Albano",
"C0""Arnaud Boutonn\303\251",
"C0""Thomas Fach-Pedersen",
"C2""Smacker video support",
"C0""Filippos Karapetis",
"",
"C1""TsAGE",
"C0""Arnaud Boutonn\303\251",
"C0""Paul Gilbert",
"",
"C1""Tucker",
"C0""Gregory Montoir",
"C2""(retired)",
"",
"C1""TwinE",
"C0""Alexandre Fontoura",
"C2""(retired)",
"C0""Vincent Hamm",
"C2""(retired)",
"C0""Felipe Sanches",
"C2""(retired)",
"C0""Nikita Tereshin",
"C2""(retired)",
"C0""Patrik Dahlstr\303\266m",
"C2""(retired)",
"C0""Arthur Blot",
"C2""(retired)",
"C0""Kyuubu",
"C2""(retired)",
"C0""To\303\253l Hartmann",
"C2""(retired)",
"C0""Seb\303\241stien Viannay",
"C2""(retired)",
"C0""Martin Gerhardy",
"",
"C1""Thimbleweed Park",
"C0""Valéry Sablonnière",
"",
"C1""Ultima",
"C0""Paul Gilbert",
"C0""Matthew Duggan",
"C0""Matthew Jimenez",
"C0""Daniel c. W\303\274rl",
"C2""(Nuvie)",
"C0""Eric Fry",
"C2""(Nuvie)",
"C0""Jeremy Newman",
"C2""(Nuvie)",
"C0""Jonathan E. Wright",
"C2""(Nuvie)",
"C0""Joseph Applegate",
"C2""(Nuvie)",
"C0""Malignant Manor",
"C2""(Nuvie)",
"C0""Markus Niemist\303\266",
"C2""(Nuvie)",
"C0""Michael Fink",
"C2""(Nuvie)",
"C0""Pieter Luteijn",
"C2""(Nuvie)",
"C0""Sam Matthews",
"C2""(Nuvie)",
"C0""Travis Howell",
"C2""(Nuvie)",
"C0""Willem Jan Palenstijn",
"C2""(Nuvie)",
"C0""Brian Tietz",
"C2""(Pentagram)",
"C0""Dominik Reichardt",
"C2""(Pentagram)",
"C0""Max Horn",
"C2""(Pentagram)",
"C0""Patrick Burke",
"C2""(Pentagram)",
"C0""Ryan Nunn",
"C2""(Pentagram)",
"C0""Willem Jan Palenstijn",
"C2""(Pentagram)",
"",
"C1""V-Cruise",
"C0""Eric Lasota",
"",
"C1""Voyeur",
"C0""Arnaud Boutonn\303\251",
"C0""Paul Gilbert",
"",
"C1""WAGE",
"C0""Alikhan Balpykov",
"C0""Eugene Sandulenko",
"",
"C1""Watchmaker",
"C0""Einar Johan Troan Somaen;",
"",
"C1""Wintermute",
"C0""Gunnar Birke",
"C2""Wintermute 3D",
"C0""Pawe\305\202 Ko\305\202odziejski",
"C2""Wintermute 3D",
"C0""Einar Johan T. S\303\270m\303\245en",
"C0""Tobia Tesan",
"C0""lolbot-iichan",
"",
"C1""Z-Vision",
"C0""Adrian Astley",
"C0""Filippos Karapetis",
"C0""Anton Yarcev",
"C0""Thomas N McEwan",
"C2""Widescreen mod, HQ panoramas, 3D audio enhancement",
"",
"",
"C1""Backend Teams",
"C1""Atari",
"C0""Miro Krop\303\241\304\215ek",
"",
"C1""Android",
"C0""Andre Heider",
"C0""Angus Lees",
"C0""Lubomyr Lisen",
"",
"C1""Dreamcast",
"C0""Marcus Comstedt",
"",
"C1""GCW0",
"C0""Eugene Sandulenko",
"",
"C1""GPH Devices (GP2X, GP2XWiz & Caanoo)",
"C0""John Willis",
"",
"C1""iPhone / iPad",
"C0""Oystein Eftevaag",
"C2""(retired)",
"C0""Vincent B\303\251nony",
"C0""Thierry Crozat",
"C0""Lars Sundstr\303\266m",
"",
"C1""LinuxMoto",
"C0""Lubomyr Lisen",
"",
"C1""Maemo",
"C0""Frantisek Dufka",
"C2""(retired)",
"C0""Tarek Soliman",
"",
"C1""Nintendo 3DS",
"C0""Thomas Edvalson",
"",
"C1""Nintendo 64",
"C0""Fabio Battaglia",
"",
"C1""Nintendo DS",
"C0""Bertrand Augereau",
"C2""HQ software scaler",
"C0""Cameron Cawley",
"C0""Neil Millstone",
"",
"C1""Nintendo Switch",
"C0""Cpasjuste",
"C0""rsn8887",
"",
"C1""OpenPandora",
"C0""John Willis",
"",
"C1""PocketPC / WinCE",
"C0""Nicolas Bacca",
"C2""(retired)",
"C0""Ismail Khatib",
"C2""(retired)",
"C0""Kostas Nakos",
"C2""(retired)",
"",
"C1""PlayStation 2",
"C0""Robert G\303\266ffringmann",
"C2""(retired)",
"C0""Max Lingua",
"",
"C1""PSP (PlayStation Portable)",
"C0""Yotam Barnoy",
"C0""Joost Peters",
"",
"C1""PlayStation Vita",
"C0""Cpasjuste",
"C0""rsn8887",
"",
"C1""SDL (Win/Linux/macOS/etc.)",
"C0""Max Horn",
"C2""(retired)",
"C0""Eugene Sandulenko",
"C2""Asm routines, GFX layers",
"",
"C1""SymbianOS",
"C0""Jurgen Braam",
"C0""Lars Persson",
"C0""Fedor Strizhniou",
"",
"C1""Tizen / BADA",
"C0""Chris Warren-Smith",
"",
"C1""Webassembly / Emscripten",
"C0""Christian K\303\274ndig",
"",
"C1""WebOS",
"C0""Klaus Reimer",
"",
"C1""Wii",
"C0""Andre Heider",
"C0""Alexander Reim",
"",
"C1""Raspberry Pi",
"C0""Manuel Alfayate",
"",
"C1""Libretro",
"C0""Giovanni Cascione",
"",
"",
"C1""Other subsystems",
"C1""Infrastructure",
"C0""Max Horn",
"C2""Backend & Engine APIs, file API, sound mixer, audiostreams, data structures, etc. (retired)",
"C0""Eugene Sandulenko",
"C0""Johannes Schickel",
"C2""(retired)",
"",
"C1""GUI",
"C0""Max Horn",
"C2""(retired)",
"C0""Vicent Marti",
"C0""Eugene Sandulenko",
"C0""Johannes Schickel",
"C2""(retired)",
"",
"C1""Miscellaneous",
"C0""David Corrales-Lopez",
"C2""Filesystem access improvements (GSoC 2007 task) (retired)",
"C0""Jerome Fisher",
"C2""MT-32 emulator",
"C0""Benjamin Haisch",
"C2""Heavily improved de-/encoder for DXA videos",
"C0""Jochen Hoenicke",
"C2""Speaker & PCjr sound support, AdLib work (retired)",
"C0""Dani\303\253l ter Laan",
"C2""Restoring original Drascula tracks, and writing convert_dxa.bat",
"C0""Chris Page",
"C2""Return to launcher, savestate improvements, leak fixes, ... (GSoC 2008 task) (retired)",
"C0""Coen Rampen",
"C2""Sound improvements",
"C0""Robin Watts",
"C2""ARM assembly routines for nice speedups on several ports; improvements to the sound mixer",
"C0""Trembyle",
"C2""Archivist",
"C0""Lothar Serra Mari",
"C2""Tackling Tremendously Tedious Tasks(tm); ScummVM's Seal of Approval (Awp?!)",
"",
"",
"C1""Website (code)",
"C0""Fredrik Wendel",
"C2""(retired)",
"",
"C1""Website (maintenance)",
"C0""James Brown",
"C2""IRC Logs maintainer",
"C0""Thierry Crozat",
"C2""Wiki maintainer",
"C0""Andre Heider",
"C2""Buildbot maintainer",
"C0""Joost Peters",
"C2""Doxygen Project Documentation maintainer",
"C0""Jordi Vilalta Prat",
"C2""Wiki maintainer",
"C0""Eugene Sandulenko",
"C2""Forum, IRC channel, Screen Shots and Mailing list maintainer",
"C0""John Willis",
"",
"C1""Website (content)",
"C0""All active team members",
"C0""",
"",
"C1""Documentation",
"C0""Thierry Crozat",
"C2""Numerous contributions to documentation",
"C0""Joachim Eberhard",
"C2""Numerous contributions to documentation (retired)",
"C0""Matthew Hoops",
"C2""Numerous contributions to documentation (retired)",
"C0""Cadi Howley",
"C2""User documentation (GSOD 2020)",
"",
"C1""Retired Team Members",
"C0""Chris Apers",
"C2""Former PalmOS porter",
"C0""Ralph Brorsen",
"C2""Help with GUI implementation",
"C0""Jamieson Christian",
"C2""iMUSE, MIDI, all things musical",
"C0""Felix Jakschitsch",
"C2""Zak256 reverse engineering",
"C0""Mutwin Kraus",
"C2""Original MacOS porter",
"C0""Peter Moraliyski",
"C2""Port: GP32",
"C0""Jeremy Newman",
"C2""Former webmaster",
"C0""Lionel Ulmer",
"C2""Port: X11",
"C0""Won Star",
"C2""Former GP32 porter",
"C0""Matan Bareket",
"C2""Website, Infrastructure, UI/UX",
"",
"",
"C1""Other contributions",
"",
"C1""Packages",
"C1""AmigaOS 4",
"C0""Hans-J\303\266rg Frieden",
"C2""(retired)",
"C0""Hubert Maier",
"C0""Juha Niemim\303\244ki",
"C2""(retired)",
"",
"C1""Atari/FreeMiNT",
"C0""Keith Scroggins",
"",
"C1""BeOS",
"C0""Stefan Parviainen",
"C2""(retired)",
"C0""Luc Schrijvers",
"",
"C1""Debian GNU/Linux",
"C0""Tore Anderson",
"C2""(retired)",
"C0""David Weinehall",
"",
"C1""Fedora / RedHat",
"C0""Willem Jan Palenstijn",
"",
"C1""Haiku",
"C0""Luc Schrijvers",
"",
"C1""macOS",
"C0""Max Horn",
"C2""(retired)",
"C0""Oystein Eftevaag",
"C2""(retired)",
"C0""Thierry Crozat",
"C0""dwa",
"C2""Tiger/Leopard PPC",
"",
"C1""Mandriva",
"C0""Dominik Scherer",
"C2""(retired)",
"",
"C1""MorphOS",
"C0""BeWorld",
"C0""Fabien Coeurjoly",
"C0""R\303\274diger Hanke",
"C2""(retired)",
"",
"C1""OS/2",
"C0""Paul Smedley",
"",
"C1""RISC OS",
"C0""Cameron Cawley",
"",
"C1""SlackWare",
"C0""Robert Kelsen",
"",
"C1""Solaris x86",
"C0""Laurent Blume",
"",
"C1""Solaris SPARC",
"C0""Markus Strangl",
"",
"C1""Win32",
"C0""Travis Howell",
"C0""Lothar Serra Mari",
"",
"C1""Win64",
"C0""Chris Gray",
"C2""(retired)",
"C0""Johannes Schickel",
"C2""(retired)",
"C0""Lothar Serra Mari",
"",
"",
"C1""GUI Translations",
"C0""Thierry Crozat",
"C2""Translation Lead",
"C1""Basque",
"C0""Mikel Iturbe Urretxa",
"",
"C1""Belarusian",
"C0""Ivan Lukyanov",
"",
"C1""Catalan",
"C0""Jordi Vilalta Prat",
"",
"C1""Czech",
"C0""Zbyn\303\254k Schwarz",
"",
"C1""Danish",
"C0""Steffen Nyeland",
"C0""scootergrisen",
"",
"C1""Dutch",
"C0""Ben Castricum",
"",
"C1""Finnish",
"C0""Linus Virtanen",
"C0""Toni Saarela",
"C0""Timo Mikkolainen",
"",
"C1""French",
"C0""Thierry Crozat",
"C0""Purple T",
"",
"C1""Galician",
"C0""Santiago G. Sanz",
"",
"C1""German",
"C0""Simon Sawatzki",
"C0""Lothar Serra Mari",
"",
"C1""Greek",
"C0""Thanasis Antoniou",
"C0""Filippos Karapetis",
"",
"C1""Hungarian",
"C0""Alex Bevilacqua",
"C0""George Kormendi",
"",
"C1""Italian",
"C0""Matteo Angelino",
"C0""Paolo Bossi",
"C0""Walter Agazzi",
"",
"C1""Norwegian (Bokm\303\245l)",
"C0""Einar Johan S\303\270m\303\245en",
"",
"C1""Norwegian (Nynorsk)",
"C0""Einar Johan S\303\270m\303\245en",
"",
"C1""Polish",
"C0""GrajPoPolsku.pl Team",
"",
"C1""Brazilian Portuguese",
"C0""ScummBR Team",
"C0""Marcel Souza Lemes",
"",
"C1""Portuguese",
"C0""Daniel Albano",
"",
"C1""Russian",
"C0""Eugene Sandulenko",
"",
"C1""Spanish",
"C0""Tom\303\241s Maidagan",
"C0""Jordi Vilalta Prat",
"C0""IlDucci",
"C0""Rodrigo Vegas S\303\241nchez-Ferrero",
"",
"C1""Swedish",
"C0""Hampus Flink",
"C0""Adrian Fr\303\274hwirth",
"",
"C1""Ukrainian",
"C0""Lubomyr Lisen",
"",
"",
"C1""Game Translations",
"C1""CGE",
"C0""Arnaud Boutonn\303\251",
"C2""Soltys English translation",
"C0""Dan Serban",
"C2""Soltys English translation",
"C0""V\303\255ctor Gonz\303\241lez",
"C2""Soltys Spanish translation",
"C0""Alejandro G\303\263mez de la Mu\303\261oza",
"C2""Soltys Spanish translation",
"",
"C1""CGE2",
"C0""Arnaud Boutonn\303\251",
"C2""Sfinx English translation",
"C0""Thierry Crozat",
"C2""Sfinx English translation",
"C0""Peter Bozs\303\263",
"C2""Sfinx English translation editor",
"C0""Ryan Clark",
"C2""Sfinx English translation editor",
"",
"C1""Drascula",
"C0""Thierry Crozat",
"C2""Improve French translation",
"",
"C1""Mortevielle",
"C0""Arnaud Boutonn\303\251",
"C2""First English translation, improve German translation",
"C0""Hugo Labrande",
"C2""Improve English translation",
"C0""Thierry Crozat",
"C2""Improve English translation",
"",
"C1""Prince",
"C0""ShinjiGR",
"C2""English translation",
"C0""Eugene Sandulenko",
"C2""English translation",
"C0""Anna Baldur",
"C2""English translation",
"",
"C1""Supernova",
"C0""Joseph-Eugene Winzer",
"C2""English translation",
"C0""Thierry Crozat",
"C2""English translation",
"C0""Walter Agazzi",
"C2""Italian translation",
"",
"",
"C1""Websites (design)",
"C0""Dob\303\263 Bal\303\241zs",
"C2""Website design",
"C0""William Claydon",
"C2""Skins for doxygen, buildbot and wiki",
"C0""Yaroslav Fedevych",
"C2""HTML/CSS for the website",
"C0""Jean Marc Gimenez",
"C2""ScummVM logo",
"C0""David Jensen",
"C2""SVG logo conversion",
"C0""Raina",
"C2""ScummVM forum buttons",
"",
"C1""Icons pack",
"C0""Stefan Philippsen",
"C2""Primary contributor and coordination",
"C0""Benjamin Funke",
"C0""Daniel Albano",
"C0""David Calvert",
"C0""Eugene Sandulenko",
"C0""J Moretti",
"C0""Jennifer McMurray",
"C0""Lothar Serra Mari",
"C0""Oleg Ermakov",
"C0""Olly Dean",
"C0""Walter Agazzi",
"C0""beresk137",
"C0""Canuma",
"C0""Thunderforge",
"C0""neuromancer",
"C0""nightm4re94",
"C0""trembyle",
"",
"C1""Code contributions",
"C0""Ori Avtalion",
"C2""Subtitle control options in the GUI; BASS GUI fixes",
"C0""Stuart Caie",
"C2""Decoders for Amiga and AtariST data files (AGOS engine)",
"C0""Paolo Costabel",
"C2""PSP port contributions",
"C0""Martin Doucha",
"C2""CinE engine objectification",
"C0""Thomas Fach-Pedersen",
"C2""ProTracker module player, Smacker video decoder",
"C0""Tobias Gunkel",
"C2""Sound support for C64 version of MM/Zak, Loom PCE support",
"C0""Dries Harnie",
"C2""Android port for ResidualVM",
"C0""Janne Huttunen",
"C2""V3 actor mask support, Dig/FT SMUSH audio",
"C0""Kov\303\241cs Endre J\303\241nos",
"C2""Several fixes for Simon1",
"C0""Jeroen Janssen",
"C2""Numerous readability and bugfix patches",
"C0""Keith Kaisershot",
"C2""Several Pegasus Prime patches and DVD additions",
"C0""Andreas Karlsson",
"C2""Initial port for SymbianOS",
"C0""Stefan Kristiansson",
"C2""Initial work on SDL2 support",
"C0""Claudio Matsuoka",
"C2""Daily Linux builds",
"C0""Thomas Mayer",
"C2""PSP port contributions",
"C0""Sean Murray",
"C2""ScummVM tools GUI application (GSoC 2007 task)",
"C0""n0p",
"C2""Windows CE port aspect ratio correction scaler and right click input method",
"C0""Mikesch Nepomuk",
"C2""MI1 VGA floppy patches",
"C0""Nicolas Noble",
"C2""Config file and ALSA support",
"C0""Tim Phillips",
"C2""Initial MI1 CD music support",
"C0""Quietust",
"C2""Sound support for Amiga SCUMM V2/V3 games, MM NES support",
"C0""Robert Crossfield",
"C2""Improved support for Apple II/C64 versions of MM",
"C0""Andreas R\303\266ver",
"C2""Broken Sword I & II MPEG2 cutscene support",
"C0""Edward Rudd",
"C2""Fixes for playing MP3 versions of MI1/Loom audio",
"C0""Daniel Schepler",
"C2""Final MI1 CD music support, initial Ogg Vorbis support",
"C0""Andr\303\251 Souza",
"C2""SDL-based OpenGL renderer",
"C0""Joel Teichroeb",
"C2""Android port for ResidualVM",
"C0""Tom Frost",
"C2""WebOS port contributions",
"",
"C1""FreeSCI Contributors",
"C0""Francois-R Boyer",
"C2""MT-32 information and mapping code",
"C0""Rainer Canavan",
"C2""IRIX MIDI driver and bug fixes",
"C0""Xiaojun Chen",
"C0""Paul David Doherty",
"C2""Game version information",
"C0""Vyacheslav Dikonov",
"C2""Config script improvements",
"C0""Ruediger Hanke",
"C2""Port to the MorphOS platform",
"C0""Matt Hargett",
"C2""Clean-ups, bugfixes, Hardcore QA, Win32",
"C0""Max Horn",
"C2""SetJump implementation",
"C0""Ravi I.",
"C2""SCI0 sound resource specification",
"C0""Emmanuel Jeandel",
"C2""Bugfixes and bug reports",
"C0""Dmitry Jemerov",
"C2""Port to the Win32 platform, numerous bugfixes",
"C0""Chris Kehler",
"C2""Makefile enhancements",
"C0""Christopher T. Lansdown",
"C2""Original CVS maintainer, Alpha compatibility fixes",
"C0""Sergey Lapin",
"C2""Port of Carl's type 2 decompression code",
"C0""Rickard Lind",
"C2""MT-32->GM MIDI mapping magic, sound research",
"C0""Hubert Maier",
"C2""AmigaOS 4 port",
"C0""Johannes Manhave",
"C2""Document format translation",
"C0""Claudio Matsuoka",
"C2""CVS snapshots, daily builds, BeOS and cygwin ports",
"C0""Dark Minister",
"C2""SCI research (bytecode and parser)",
"C0""Carl Muckenhoupt",
"C2""Sources to the SCI resource viewer tools that started it all",
"C0""Anders Baden Nielsen",
"C2""PPC testing",
"C0""Walter van Niftrik",
"C2""Ports to the Dreamcast and GP32 platforms",
"C0""Rune Orsval",
"C2""Configuration file editor",
"C0""Solomon Peachy",
"C2""SDL ports and much of the sound subsystem",
"C0""Robey Pointer",
"C2""Bug tracking system hosting",
"C0""Magnus Reftel",
"C2""Heap implementation, Python class viewer, bugfixes",
"C0""Christoph Reichenbach",
"C2""UN*X code, VM/Graphics/Sound/other infrastructure",
"C0""George Reid",
"C2""FreeBSD package management",
"C0""Lars Skovlund",
"C2""Project maintenance, most documentation, bugfixes, SCI1 support",
"C0""Rink Springer",
"C2""Port to the DOS platform, several bug fixes",
"C0""Rainer De Temple",
"C2""SCI research",
"C0""Sean Terrell",
"C0""Hugues Valois",
"C2""Game selection menu",
"C0""Jordi Vilalta",
"C2""Numerous code and website clean-up patches",
"C0""Petr Vyhnak",
"C2""The DCL-IMPLODE algorithm, many Win32 improvements",
"C0""Bas Zoetekouw",
"C2""Man pages, debian package management, CVS maintenance",
"C0""Special thanks to Prof. Dr. Gary Nutt for allowing the FreeSCI VM extension as a course project in his Advanced OS course.",
"C0""",
"C0""Special thanks to Bob Heitman and Corey Cole for their support of FreeSCI.",
"C0""",
"",
"C1""ResidualVM Contributors",
"C1""Grim",
"C0""Thomas Allen",
"C2""Various engine code fixes and improvements",
"C0""Torbj\303\266rn Andersson",
"C2""Various code fixes",
"C0""Ori Avtalion",
"C2""Lipsync, LAF support, various code fixes",
"C0""Robert Biro",
"C2""Antialiasing support",
"C0""Bastien Bouclet",
"C2""Various fixes to engine",
"C0""David Cardwell",
"C2""Few fixes to EMI",
"C0""Marcus Comstedt",
"C2""Initial Dreamcast port",
"C0""Andrea Corna",
"C2""Patcher module, various engine improvements",
"C0""Jonathan Gray",
"C2""Various code fixes",
"C0""Tobias Gunkel",
"C2""Initial Android port, few engines fixes",
"C0""Azamat H. Hackimov",
"C2""Configure fix",
"C0""Vincent Hamm",
"C2""Various engine code fixes and improvements",
"C0""Sven Hesse",
"C2""Various compilation fixes",
"C0""Matthew Hoops",
"C2""Smush codec48, Grim and EMI engine improvements",
"C0""Erich Hoover",
"C2""x86-64 fixes, various code fixes and improvements",
"C0""Max Horn",
"C2""Few code fixes",
"C0""Travis Howell",
"C2""Various code fixes, Windows port",
"C0""Joseph Jezak",
"C2""A lot of engine improvements and fixes",
"C0""Guillem Jover",
"C2""Few code improvements",
"C0""Filippos Karapetis",
"C2""Compilation fixes",
"C0""Ingo van Lil",
"C2""Various fixes and improvements for EMI",
"C0""Vincent Pelletier",
"C2""Various engine and TinyGL improvements",
"C0""Joost Peters",
"C2""Various code fixes",
"C0""George Macon",
"C2""Few fixes",
"C0""Josh Matthews",
"C2""Few fixes to engine",
"C0""Matthieu Milan",
"C2""Various engine improvements",
"C0""Gregory Montoir",
"C2""Few fixes to engine",
"C0""Stefano Musumeci",
"C2""TinyGL backend and engine driver improvements",
"C0""Christian Neumair",
"C2""Various optimisation patches",
"C0""Daniel Schepler",
"C2""Initial grim engine contributor, LUA support",
"C0""Dmitry Smirnov",
"C2""Minor spelling corrections",
"C0""Yaron Tausky",
"C2""Fixes to subtitles",
"C0""Julien Templier",
"C2""create_project tool",
"C0""Pino Toscano",
"C2""Debian GNU/Linux package files",
"C0""Lionel Ulmer",
"C2""OpenGL optimisations",
"C0""cmayer0087",
"C2""Various engine code fixes",
"C0""JenniBee",
"C2""Compilation fixes",
"C0""karjonas",
"C2""Various engine code fixes",
"C0""mparnaudeau",
"C2""Various grim engine code fixes",
"C0""PoulpiFr",
"C2""Few fixes to Android port",
"C0""sietschie",
"C2""Few fixes to engine",
"",
"C1""Myst 3",
"C0""David Fioramonti",
"C2""Autosave support and few fixes",
"C0""Matthew Hoops",
"C2""Various engine improvements and code fixes",
"C0""Stefano Musumeci",
"C2""TinyGL engine support",
"",
"C1""Stark",
"C0""Bartosz Dudziak",
"C2""Various engine improvements and code fixes",
"C0""Matthew Hoops",
"C2""ADPCM decoder",
"C0""Pawe\305\202 Ko\305\202odziejski",
"C2""Various engine code fixes",
"C0""Awad Mackie",
"C2""Few fixes to engine",
"C0""Marius Ioan Orban",
"C2""Code fix",
"C0""Vincent Pelletier",
"C2""Raw sound support",
"C0""Jordi Vilalta Prat",
"C2""Initial engine contributor",
"C0""Scott Thomas",
"C2""Initial engine author",
"C0""Will Thomson",
"C2""Few fixes to engine",
"C0""Faalagorn",
"C2""Few code improvements",
"C0""orangeforest11",
"C2""Few engine improvements",
"",
"",
"C0""And to all the contributors, users, and beta testers we've missed. Thanks!",
"C0""",
"",
"C1""Special thanks to",
"",
"C0""Daniel Balsom",
"C2""For the original Reinherit (SAGA) code",
"C0""Sander Buskens",
"C2""For his work on the initial reversing of Monkey2",
"C0""Dean Beeler",
"C2""For the original MT-32 emulator",
"C0""Kevin Carnes",
"C2""For Scumm16, the basis of ScummVM's older gfx codecs",
"C0""Curt Coder",
"C2""For the original TrollVM (preAGI) code",
"C0""Patrick Combet",
"C2""For the original Gobliiins ADL player",
"C0""Ivan Dubrov",
"C2""For contributing the initial version of the Gobliiins engine",
"C0""Henrik Engqvist",
"C2""For generously providing hosting for our buildbot, SVN repository, planet and doxygen sites as well as tons of HD space",
"C0""DOSBox Team",
"C2""For their awesome OPL2 and OPL3 emulator",
"C0""Yusuke Kamiyamane",
"C2""For contributing some GUI icons",
"C0""Till Kresslein",
"C2""For design of modern ScummVM GUI",
"C0""Jezar Wakefield",
"C2""For his freeverb filter implementation",
"C0""Jim Leiterman",
"C2""Various info on his FM-TOWNS/Marty SCUMM ports",
"C0""Lloyd Rosen",
"C2""For deep tech details about C64 Zak & MM",
"C0""Sarien Team",
"C2""Original AGI engine code",
"C0""Jimmi Th\303\270gersen",
"C2""For ScummRev, and much obscure code/documentation",
"C0""Tristan Matthews",
"C2""For additional work on the original MT-32 emulator",
"C0""James Woodcock",
"C2""Soundtrack enhancements",
"C0""Anton Yartsev",
"C2""For the original re-implementation of the Z-Vision engine",
"C0""Tony Warriner and everyone at Revolution Software Ltd. for sharing with us the source of some of their brilliant games, allowing us to release Beneath a Steel Sky as freeware... and generally being supportive above and beyond the call of duty.",
"C0""",
"C0""John Passfield and Steve Stamatiadis for sharing the source of their classic title, Flight of the Amazon Queen and also being incredibly supportive.",
"C0""",
"C0""Joe Pearce from The Wyrmkeep Entertainment Co. for sharing the source of their famous title Inherit the Earth, for sharing the source of The Labyrinth of Time and for always replying promptly to our questions.",
"C0""",
"C0""Aric Wilmunder, Ron Gilbert, David Fox, Vince Lee, and all those at LucasFilm/LucasArts who made SCUMM the insane mess to reimplement that it is today. Feel free to drop us a line and tell us what you think, guys!",
"C0""",
"C0""Alan Bridgman, Simon Woodroffe and everyone at Adventure Soft for sharing the source code of some of their games with us.",
"C0""",
"C0""John Young, Colin Smythe and especially Terry Pratchett himself for sharing the source code of Discworld I & II with us.",
"C0""",
"C0""Emilio de Paz Arag\303\263n from Alcachofa Soft for sharing the source code of Drascula: The Vampire Strikes Back with us and his generosity with freewaring the game.",
"C0""",
"C0""David P. Gray from Gray Design Associates for sharing the source code of the Hugo trilogy.",
"C0""",
"C0""The mindFactory team for writing Broken Sword 2.5, a splendid fan-made sequel, and for sharing the source code with us.",
"C0""",
"C0""Neil Dodwell and David Dew from Creative Reality for providing the source of Dreamweb and for their tremendous support.",
"C0""",
"C0""Janusz Wi\305\233niewski and Miroslaw Liminowicz from Laboratorium Komputerowe Avalon for providing full source code for So\305\202tys and Sfinx and letting us redistribute the games.",
"C0""",
"C0""Jan Nedoma for providing the sources to the Wintermute-engine, and for his support while porting the engine to ScummVM.",
"C0""",
"C0""Bob Bell, David Black, Michel Kripalani, and Tommy Yune from Presto Studios for providing the source code of The Journeyman Project: Pegasus Prime and The Journeyman Project 2: Buried in Time.",
"C0""",
"C0""Electronic Arts IP Preservation Team, particularly Stefan Serbicki, and Vasyl Tsvirkunov of Electronic Arts for providing the source code of the two Lost Files of Sherlock Holmes games. James M. Ferguson and Barry Duncan for their tenacious efforts to recover the sources.",
"C0""",
"C0""John Romero for sharing the source code of Hyperspace Delivery Boy! with us.",
"C0""",
"C0""Steffen Dingel for sharing the source code of the Mission Supernova game with us.",
"C0""",
"C0""The LUA developers, for creating a nice compact script interpreter.",
"C0""",
"C0""Tim Schafer, for obvious reasons, and everybody else who helped make Grim Fandango a brilliant game; and the EMI team for giving it their best try.",
"C0""",
"C0""Bret Mogilefsky, for managing to create a SPUTM-style 3D LUA engine, and avoiding the horrible hack it could have been.",
"C0""",
"C0""Benjamin Haisch, for emimeshviewer, which our EMI code borrows heavily from.",
"C0""",
"C0""Fabrizio Lagorio from Trecision S.p.A., for finding and providing the source code of many of their games.",
"C0""",
"C0""Ron Davis for releasing the sources and generously giving away the three chapters of God of Thunder.",
"C0""",
"",
};
