/* ScummVM - Graphic Adventure Engine
 *
 * ScummVM is the legal property of its developers, whose names
 * are too numerous to list here. Please refer to the COPYRIGHT
 * file distributed with this source distribution.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * aint32 with this program; if not, write to the Free Software
 *
 *
 * Based on the original sources
 *   Faery Tale II -- The Halls of the Dead
 *   (c) 1993-1996 The Wyrmkeep Entertainment Co.
 */

#ifndef SAGA2_PROGRAM_H
#define SAGA2_PROGRAM_H

namespace Saga2 {

#define PROGRAM_BRIEF_NAME      "FTA2"
#define PROGRAM_FULL_NAME       "Faery Tale Adventure II -- The Halls of the Dead"
#define PROGRAM_COPYRIGHT       "Copyright (c) 1996 by The Wyrmkeep Entertainment Co."
#define PROGRAM_CONFIG_FILE     PROGRAM_BRIEF_NAME".INI"

#define MUTEX_NAME              "FTA2Mutex"

/* ===================================================================== *
    Version Stamp values
 * ===================================================================== */

#define PROGRAM_COMPANY "The Wyrmkeep Entertainment Co.\0"
#define PROGRAM_DESCRIP "Faery Tale Adventure II - The  Halls of the Dead\0"

#define PROGRAM_MODULEN "Fta2Win\0"
#define PROGRAM_COPYRGT "Copyright � 1996 The Wyrmkeep Entertainment Co.\0"
#define PROGRAM_EXENAME "FTA2.EXE\0"
#define PROGRAM_PRODUCT "Faery Tale II - Halls of the Dead\0"

} // end of namespace Saga2

#endif
