/* ScummVM - Graphic Adventure Engine
 *
 * ScummVM is the legal property of its developers, whose names
 * are too numerous to list here. Please refer to the COPYRIGHT
 * file distributed with this source distribution.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef TWP_OBJECT_ANIM_H
#define TWP_OBJECT_ANIM_H

#include "common/str.h"
#include "common/str-array.h"
#include "common/array.h"
#include "math/vector2d.h"

namespace Twp {
struct ObjectAnimation {
	Common::String name;
	Common::String sheet;
	Common::StringArray frames;
	Common::Array<ObjectAnimation> layers;
	Common::StringArray triggers;
	Common::Array<Math::Vector2d> offsets;
	bool loop = false;
	float fps = 0.f;
	int flags = 0;
	int frameIndex = 0;
};
} // namespace Twp

#endif
