/* ScummVM - Graphic Adventure Engine
 *
 * ScummVM is the legal property of its developers, whose names
 * are too numerous to list here. Please refer to the COPYRIGHT
 * file distributed with this source distribution.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef AGS_ENGINE_AC_SCREEN_H
#define AGS_ENGINE_AC_SCREEN_H

#include "ags/engine/ac/dynobj/script_user_object.h"

namespace AGS3 {

int  Screen_GetScreenWidth();
int  Screen_GetScreenHeight();
bool Screen_GetAutoSizeViewport();
void Screen_SetAutoSizeViewport(bool on);
int  Screen_GetViewportCount();
ScriptViewport *Screen_GetViewport();
ScriptViewport *Screen_GetAnyViewport(int index);
ScriptUserObject *Screen_RoomToScreenPoint(int roomx, int roomy);

namespace AGS {
namespace Shared {
class Bitmap;
} // namespace Shared
} // namespace AGS
namespace AGS {
namespace Engine {
class IDriverDependantBitmap;
} // namespace Engine
} // namespace AGS

void fadein_impl(PALETTE p, int speed);
void fadeout_impl(int spdd);
void current_fade_out_effect();
AGS::Engine::IDriverDependantBitmap *prepare_screen_for_transition_in(bool opaque);

} // namespace AGS3

#endif
