/* ScummVM - Graphic Adventure Engine
 *
 * ScummVM is the legal property of its developers, whose names
 * are too numerous to list here. Please refer to the COPYRIGHT
 * file distributed with this source distribution.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef HODJNPODJ_GARFUNKLE_RESOURCE_H
#define HODJNPODJ_GARFUNKLE_RESOURCE_H

namespace Bagel {
namespace HodjNPodj {
namespace Garkfunkle {

#define IDC_SCROLL                      114

#define IDD_SUBOPTIONS                  115

#define IDD_RULES_DIALOG                900

#define IDC_RULES                       1002

#define IDC_NUMBUTTONS                  1021
#define IDC_SPEED                       1022
#define IDC_PLAYGAME                    1023
#define IDC_PLAYMUSIC                   1024

} // namespace Garfunkle
} // namespace HodjNPodj
} // namespace Bagel

#endif
