/* ScummVM - Graphic Adventure Engine
 *
 * ScummVM is the legal property of its developers, whose names
 * are too numerous to list here. Please refer to the COPYRIGHT
 * file distributed with this source distribution.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef HODJNPODJ_LIFE_RESOURCE_H
#define HODJNPODJ_LIFE_RESOURCE_H

namespace Bagel {
namespace HodjNPodj {
namespace Life {

#define LIFE_ICON                       2
//#define IDD_COMMANDS                    5
//#define IDD_RULES                       111
//#define IDD_STATS                       114
//#define IDD_NEWSETTING                  118
//#define IDI_LIFE                        123
//#define IDD_RULES_DIALOG                900
//#define IDC_RULES_OKAY                  900
//#define IDC_RULES_ARROWDN               910
//#define IDC_RULES_ARROWUP               911
//#define IDC_RULES_INVALID               912
//#define IDC_RULES                       1001
//#define IDC_NEWGAME                     1004
//#define IDC_OPTIONS                     1005
//#define IDC_QUIT                        1007

#define IDD_USERCFG                     110     // user config dialog id
#define IDC_STARTVILLAGES               1010    // scroll bar 
#define IDC_EVOLVESPEED                 1011    // scroll bar
#define IDC_SCORE                       1013
#define IDC_YEAR                        1014
#define IDC_COL_LEFT                    1015

#define IDE_SCORE                       1016
#define IDE_YEAR                        1017
#define IDE_TIMER                       1018
#define IDE_COL_LEFT                    1022

#define IDC_EVOLVEROUNDS                1019
#define IDC_RETURN_LIFE                 1020
#define IDC_ROUNDS                      1025
#define IDC_PREPLACE                    1027

} // namespace Life
} // namespace HodjNPodj
} // namespace Bagel

#endif
