/* ScummVM - Graphic Adventure Engine
 *
 * ScummVM is the legal property of its developers, whose names
 * are too numerous to list here. Please refer to the COPYRIGHT
 * file distributed with this source distribution.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef BAGEL_METAGAME_FRAME_RESOURCE_H
#define BAGEL_METAGAME_FRAME_RESOURCE_H

namespace Bagel {
namespace HodjNPodj {
namespace Metagame {
namespace Frame {

//{{NO_DEPENDENCIES}}
// App Studio generated include file.
// Used by HODJPODJ.RC
//
#define IDI_POKER                       2
#define IDB_MAP                         248
#define IDI_ICON1                       249
#define IDI_CESDEMO                     249
#define IDD_MAIN_SCROLL                 250
#define IDD_META_SETUP                  251
#define IDD_RESTORE                     252
#define IDC_PLAY_META                   1022
#define IDC_RESTORE_GAME                1023
#define IDC_PLAY                        1023
#define IDC_PLAY_MINI                   1024
#define IDC_CANCEL_PLAY                 1024
#define IDC_GRAND_TOUR                  1025
#define IDC_HODJ_SKILL_HARD             1025
#define IDC_RESTART_MOVIE               1026
#define IDC_HODJ_SKILL_MEDIUM           1026
#define IDC_HODJ_SKILL_EASY             1027
#define IDC_QUIT_GAME                   1028
#define IDC_PODJ_SKILL_EASY             1028
#define IDC_PODJ_SKILL_MEDIUM           1029
#define IDC_PODJ_SKILL_HARD             1030
#define IDC_GAMETIME_LONG               1031
#define IDC_GAMETIME_MEDIUM             1032
#define IDC_GAMETIME_SHORT              1033
#define IDC_PODJ_COMPUTER               1034
#define IDC_PODJ_HUMAN                  1035
#define IDC_SAVEGAME                    1036
#define IDC_PLAY_HODJ_MOVIE             1037
#define IDC_PLAY_PODJ_MOVIE             1038
#define IDC_PLAY_LOGO_MOVIE             1039
#define IDC_PLAY_TITLE_MOVIE            1040
#define IDC_RSLOT1                      1040
#define IDC_RSLOT2                      1041
#define IDC_RSLOT3                      1042
#define IDC_RSLOT4                      1043
#define IDC_RSLOT5                      1044
#define IDC_RSLOT6                      1045
#define IDC_SHOW_CREDITS                1046

} // namespace Frame
} // namespace Metagame
} // namespace HodjNPodj
} // namespace Bagel

#endif
